/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *       CLASS FILELISTENER                               * */
/* *                                                        * */
/* *  $Module:   FILELISTENER                               * */ 
/* *                                                        * */
/* *  Copyright (C) 2001, 2002 MPI fuer Informatik          * */
/* *                                                        * */
/* *  This program is free software; you can redistribute   * */
/* *  it and/or modify it under the terms of the GNU        * */
/* *  General Public License as published by the Free       * */
/* *  Software Foundation; either version 2 of the License, * */
/* *  or (at your option) any later version.                * */
/* *                                                        * */
/* *  This program is distributed in the hope that it will  * */
/* *  be useful, but WITHOUT ANY WARRANTY; without even     * */
/* *  the implied warranty of MERCHANTABILITY or FITNESS    * */
/* *  FOR A PARTICULAR PURPOSE.  See the GNU General Public * */
/* *  License for more details.                             * */
/* *                                                        * */
/* *  You should have received a copy of the GNU General    * */
/* *  Public License along with this program; if not, write * */
/* *  to the Free Software Foundation, Inc., 59 Temple      * */
/* *  Place, Suite 330, Boston, MA  02111-1307  USA         * */
/* *                                                        * */
/* *                                                        * */
/* $Revision: 1.1.1.1 $                                         * */
/* $State: Exp $                                            * */
/* $Date: 2006/02/20 10:28:30 $                             * */
/* $Author: topic $                                         * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Christoph Weidenbach                       * */
/* *             MPI fuer Informatik                        * */
/* *             Im Stadtwald                               * */
/* *             66123 Saarbruecken                         * */
/* *             Email: weidenb@mpi-inf.mpg.de               * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/

/**************************************************************/
/*                         INCLUDES                           */
/**************************************************************/

#include <qdir.h>
#include <qmessagebox.h>
#include <qstring.h>
//Added by qt3to4:
#include <QFileDialog>
#include <QTextStream>

#include "FileListener.h"
#include "SPASSdefines.h"
#include "SPASSDialogs.h"

/**************************************************************/
/*                        Constants                           */
/**************************************************************/

/* The default problem file name filter (*.dfg) */
const QString FileListener::DFG_FILTER("DFG problem files (*.dfg)");

/**************************************************************/
/*                     Static Fields                          */
/**************************************************************/

/* The start directory for file dialogs.
 * We have to manage the memory for this one ourself.
 */
QString FileListener::current_dir = QDir::homePath();

/**************************************************************/
/*                        Constructors                        */
/**************************************************************/
/**********************************************************
  INPUT:   SPASS files, a splitter, and the status bar.
  RETURNS: None.
  EFFECT:  Creates an instance of this class.
***********************************************************/
FileListener::FileListener(SPASSFiles *f,
			   SPASSSplitter *splitter,
			   SPASSStatusBar *status)
  : Listener()
{
  files = f;
  splitterWidget = splitter;
  statusbar = status;
}


/**************************************************************/
/*                        protected methods                   */
/**************************************************************/

/**********************************************************
  INPUT:   Widget title, and a mode.
  RETURNS: None.
  EFFECT:  Asks the user to select a file name and returns
           the name of the selected file. If the user
           doesn't select a file, returns an empty string.
***********************************************************/
QString FileListener::getFileName(const QString & widget_title,
				  QFileDialog::FileMode mode)
{
  QString file_name;
  QFileDialog * file_dialog = getExecutedFileDialog(widget_title,
						    mode);

  /* If the user rejected to select a file name,
   * return "".
   * Else return selected file name.
   */
  if (file_dialog->result() == QDialog::Rejected) {

    /* no file selected, so return an empty string */
    file_name = "";
  }
  else {
    
      /* return the selected file */
      file_name = file_dialog->selectedFile();
  }

  delete file_dialog;

  return file_name;
}

/**********************************************************
  INPUT:   QString filename
  RETURNS: None.
  EFFECT:  handles the opening of an input file.
***********************************************************/
void FileListener::handleFileOpen(QString filename)
{
  if (!filename.isEmpty()) {
    delete files->input->file;
    delete files->input->fileinfo;

    files->input->file = new QFile(filename);
    Q_CHECK_PTR(files->input->file);
    files->input->fileinfo = new QFileInfo(files->input->file->name());
    Q_CHECK_PTR(files->input->fileinfo);

    /* Open file in different modes depending on the user's rights */
    if (files->input->file->open(QIODevice::ReadOnly)) {
      showFile();
      files->input->file->close();
    }
    else {
      SPASSDialogs::showErrorMessage (ERR_FILE_NOT_READABLE);
    }
  }
}

/**********************************************************
  INPUT:   None.
  RETURNS: None.
  EFFECT:  handles the saving of a modified input file.
***********************************************************/
int FileListener::handleModifiedInput()
{
  int answer = QMessageBox::Yes;

  // check if current file needs to be saved
  if (files->input->file != NULL) {
    if (splitterWidget->getInput()->isModified()) {
      answer = SPASSDialogs::showWarningToSave(files);
      if (files->input->fileinfo->isWritable() && (answer == QMessageBox::Yes)) {
	save();
      }
    }
  }

  return answer;
}

/**********************************************************
  INPUT:   None.
  RETURNS: None.
  EFFECT:  Saves the contents of the editor in the given
           file, making it the new input file. Additionally,
           it diplays the name of the new input file in the
           status bar.
***********************************************************/
void FileListener::save(void)
{
    save(files->input->file->name());
}

/**********************************************************
  INPUT:   A file name.
  RETURNS: None.
  EFFECT:  Saves the contents of the editor in the given
           file, making it the new input file. Additionally,
           it diplays the name of the new input file in the
           status bar.
***********************************************************/
void FileListener::save(const char *fname)
{

  /* Don't attempy to save files with an empty name */  
  if (!strcmp(fname, "")) {
    return;
  }

  QString filename(fname);

  delete files->input;
  files->input = new SPASSFiles::Input(filename);
  Q_CHECK_PTR(files->input);
  
  files->input->file->open(QIODevice::WriteOnly);
  QTextStream stream(files->input->file);
  stream << splitterWidget->getInput()->text();
  files->input->file->close();
    
  statusbar->showFileName(files->input->file->name());
}

/**********************************************************
  INPUT:   None.
  RETURNS: None.
  EFFECT:  shows the input file in the fileeditor.
***********************************************************/
void FileListener::showFile()
{
  QTextStream stream(files->input->file);

  statusbar->showFileName(QString(files->input->file->name()));

  splitterWidget->getInput()->clear();
  splitterWidget->getInput()->setText(stream.read());
  splitterWidget->getInput()->setModified(false);
}

/**************************************************************/
/*                        private methods                   */
/**************************************************************/
/**********************************************************
  INPUT:   A file dialog mode.
  RETURNS: An executed file dialog.
  EFFECT:  Handles the setting of current directory, builds
           a nice file dialog and executes it.
***********************************************************/
QFileDialog* FileListener::getExecutedFileDialog(const QString & widget_title,
						 QFileDialog::Mode mode) {
   
  /* create fancy file dialog */
  QFileDialog* file_dialog = new QFileDialog(NULL, widget_title);
  file_dialog->setFilter(DFG_FILTER);
  file_dialog->setMode(mode);
  //  file_dialog->setCaption(widget_title);
  
  /* set the current directory.
   * If none is set, use the home directory
   */
    file_dialog->setDir(current_dir);
 
  /* execute the file dialog.
   * It's a modal dialog, so it's executed,
   * not shown.
   */
  file_dialog->exec();

  /* store user's current directory for next call */
  setCurrentDir(file_dialog->directory().absolutePath());

  return file_dialog;
}

/**********************************************************
  INPUT:   A directory.
  RETURNS: None.
  EFFECT:  Sets the current directory, and deletes the
           previous one, if there was one.
***********************************************************/
void FileListener::setCurrentDir(const QString & dir) {

  /* save the current directiory */
  current_dir = dir;
}
