/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *           CLASS RUNAPPLICATIONLISTENER                 * */
/* *                                                        * */
/* *  $Module:   RUNAPPLICATIONLISTENER                     * */ 
/* *                                                        * */
/* *  Copyright (C) 2001, 2002 MPI fuer Informatik          * */
/* *                                                        * */
/* *  This program is free software; you can redistribute   * */
/* *  it and/or modify it under the terms of the GNU        * */
/* *  General Public License as published by the Free       * */
/* *  Software Foundation; either version 2 of the License, * */
/* *  or (at your option) any later version.                * */
/* *                                                        * */
/* *  This program is distributed in the hope that it will  * */
/* *  be useful, but WITHOUT ANY WARRANTY; without even     * */
/* *  the implied warranty of MERCHANTABILITY or FITNESS    * */
/* *  FOR A PARTICULAR PURPOSE.  See the GNU General Public * */
/* *  License for more details.                             * */
/* *                                                        * */
/* *  You should have received a copy of the GNU General    * */
/* *  Public License along with this program; if not, write * */
/* *  to the Free Software Foundation, Inc., 59 Temple      * */
/* *  Place, Suite 330, Boston, MA  02111-1307  USA         * */
/* *                                                        * */
/* *                                                        * */
/* $Revision: 1.1.1.1 $                                         * */
/* $State: Exp $                                            * */
/* $Date: 2006/02/20 10:28:30 $                             * */
/* $Author: topic $                                         * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Christoph Weidenbach                       * */
/* *             MPI fuer Informatik                        * */
/* *             Im Stadtwald                               * */
/* *             66123 Saarbruecken                         * */
/* *             Email: weidenb@mpi-inf.mpg.de               * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/

/**************************************************************/
/*                         INCLUDES                           */
/**************************************************************/
#include <qmessagebox.h>

#include "KillApplicationListener.h"
#include "RunApplicationListener.h"
#include "SPASSDialogs.h"

/**************************************************************/
/*                        Constructors                        */
/**************************************************************/
/**********************************************************
  INPUT:   A process, an application name, an options input
           widget, a tab widget, an output widget, the
           status bar, and an action.
  RETURNS: None.
  EFFECT:  Creates an instance of this class.
***********************************************************/
RunApplicationListener::RunApplicationListener(Q3Process* proc,
					       const QString & app,
					       QComboBox * opt,
					       SPASSSplitter * splitter,
					       SPASSStatusBar* status,
					       QAction* action)
  : Listener()
{
  process               = proc;
  application           = app;
  optionsWidget         = opt;
  splitterWidget        = splitter;
  statusbar             = status;

  /* Connect signals and slots. */
  connect(action, SIGNAL(activated()),
	  this, SLOT(listen()));
  connect(process, SIGNAL(readyReadStdout()),
	  this, SLOT(readFromSubprocessStdout()));
  connect(process, SIGNAL(readyReadStderr()),
	  this, SLOT(readFromSubprocessStderr()));
}

/**************************************************************/
/*                      Deconstructors                        */
/**************************************************************/
/**********************************************************
  INPUT:   None.
  RETURNS: None.
  EFFECT:  Deletes this object. It kills the associated 
           process. 
***********************************************************/
RunApplicationListener::~RunApplicationListener(void)
{
  KillApplicationListener::killProcess(process);
}

/**************************************************************/
/*                          Slots                             */
/**************************************************************/
/**********************************************************
  INPUT:   None.
  RETURNS: None.
  EFFECT:  Launches the subprocess.
***********************************************************/
void RunApplicationListener::listen(void)
{
  if (preprocess()){ 
    splitterWidget->getOutput()->clear();

    if (!process->start()) {
      SPASSDialogs::showErrorMessage("Unable to launch subprocess.\n\n"
				     "Make sure that all your settings are correct.\n"
				     "A typical mistake would be a PATH environment "
				     "without the directory of SPASS executables");
    }
  }
}

/**********************************************************
  INPUT:   None.
  RETURNS: None.
  EFFECT:  Reads from subprocess stdout.
***********************************************************/
void RunApplicationListener::readFromSubprocessStdout(void)
{
  splitterWidget->getOutput()->insert(process->readStdout());
}

/**********************************************************
  INPUT:   None.
  RETURNS: None.
  EFFECT:  Reads from subprocess stderr.
***********************************************************/
void RunApplicationListener::readFromSubprocessStderr(void)
{
  SPASSDialogs::showErrorMessage("Subprocess error:\n" 
				 + QString(process->readStderr()) 
				 + "\n");
}

/**************************************************************/
/*                        protected methods                   */
/**************************************************************/
/**********************************************************
  INPUT:   None.
  RETURNS: True if a new process should be started, false
           otherwise.
  EFFECT:  If a process is running, asks the user if it
           should be killed. Sets the process arguments.
***********************************************************/

bool RunApplicationListener::preprocess(void)
{
  //If a process is running already, ask the user if she wants to kill it.
  if (process->isRunning()) {
    if (SPASSDialogs::showConfirmation("Do you want to kill the running "
				       + application
				       + "?")
	== QMessageBox::Yes) {
      // kill the process. hanging up is safer, but won't always work.
      process->kill();
    }
    else {
      // keep the process running.
      return false;
    }
  }

  process->setArguments(createArgumentList());
  return true;
}

/**************************************************************/
/*                        private methods                     */
/**************************************************************/
/**********************************************************
  INPUT:   None.
  RETURNS: The string list of arguments..
  EFFECT:  Creates a list of arguments from application
           name, options and a temporary input file name.
***********************************************************/
QStringList RunApplicationListener::createArgumentList(void)
{
  return QStringList::split(" ",
			    application 
			    + " " 
			    + (optionsWidget == NULL ? QString("") : optionsWidget->currentText())
			    + " " 
			    + splitterWidget->getInput()->getTemporaryFileName());
}






